"""Celery tasks: post log to cloudwatch"""
import json  # type: ignore

import celery  # type: ignore
import redis  # type: ignore
from celery import shared_task  # type: ignore
from django.utils import timezone  # type: ignore

import app_settings as app_settings
from cloudwatch import CloudWatchLog

r = redis.StrictRedis()
pipe = r.pipeline()


def store_to_redis(data, name='logbacklog'):
    """
    Stores a python dictionary to the specified redis queue
    :param data: a python dictionary
    :param name: name of the redis queue
    :return: None
    """
    json_data = json.dumps(data)
    pipe.rpush(name, json_data)
    pipe.execute()


def fetch_from_redis(name='logbacklog'):
    """
    Loads data from redis and turns it into a (python) list of dicts
    :param name:
    :return: List of python dictionaries
    """
    pipe.lrange(name, 0, -1)
    redis_data = pipe.execute()
    data = []
    for entry in redis_data[0]:
        data.append(json.loads(entry))
    return data


class LogToCloudWatchBase(celery.Task):
    """Create log object"""
    _cw = None

    @property
    def cw(self):
        """Get cw-log object"""
        if self._cw is None:
            self._cw = CloudWatchLog()
        return self._cw


@shared_task
def log_to_redis(**log_data):
    """Store log in redis"""

    try:
        store_to_redis(log_data)
        log_from_cloudwatch.delay()

    except Exception as e:
        print('Exception', e)
        log_to_redis.retry(
            countdown=2 ** log_to_redis.request.retries, exc=e,
        )


@celery.task(base=LogToCloudWatchBase, max_retries=10, rate_limit=5)
def log_from_cloudwatch(name='logbacklog'):
    """Create log or display error"""

    try:
        # Logs should be send as groups and only after a set interval
        # the interval is demarcated by the 'queue_occupied' value on redis
        if not r.get('queue_occupied'):
            log_data = fetch_from_redis()

            # A post command should not contain empty data, nor
            # should a deletion command be triggered before the logs
            # have been pushed to cloudwatch
            if len(log_data) > 0:
                log_from_cloudwatch.cw.put_logs(log_data)
                pipe.delete(name)

                # Occupy the queue untill the interval has passed.
                pipe.set(
                    'queue_occupied', timezone.now(),
                    ex=app_settings.LOG_INTERVAL,
                )
                pipe.execute()

    except Exception as e:
        print('Exception', e)
        log_from_cloudwatch.retry(
            countdown=2 ** log_from_cloudwatch.request.retries, exc=e,
        )



"""App specific settings"""
from django.conf import settings  # type: ignore

AUDIT_LOGGER_REGION = getattr(settings, 'AUDIT_LOG_REGION', 'eu-west-1')
AUDIT_LOGGER_KEY_ID = getattr(settings, 'AWS_ACCESS_KEY_ID', None)
AUDIT_LOGGER_SECRET_ACCESS_KEY = getattr(
    settings, 'AWS_SECRET_ACCESS_KEY', None,
)
AUDIT_LOGGER_GROUP_NAME = getattr(settings, 'AUDIT_LOG_GROUP_NAME', None)
AUDIT_LOGGER_STREAM_NAME = getattr(settings, 'AUDIT_LOG_STREAM_NAME', None)
AUDIT_LOGGER_EXCLUDE = getattr(settings, 'AUDIT_LOG_EXCLUDE', [])
# Interval value represents minutes
LOG_INTERVAL = getattr(settings, 'LOG_INTERVAL', 5)



"""
Tests for module

- Make sure the decorator is working
- Make sure non-decorated models are not being logged
"""
from django.contrib.auth.models import User  # type: ignore
from django.db import models  # type: ignore
from django.test import TestCase  # type: ignore
from django.test.utils import override_settings  # type: ignore

from signals import auditlogger_update_delete  # type: ignore

# build in try/except for python2
try:
    from unittest.mock import patch as patch  # noqa
except ImportError:
    from mock import patch as patch  # type: ignore  # noqa


@auditlogger_update_delete
class LoggedTestModel(models.Model):
    """Test model that is being logged"""
    name = models.CharField(max_length=100, default='LoggedTestModel')


class TestModel(models.Model):
    """Test model that is not being logged"""
    name = models.CharField(max_length=100, default='TestModel')


class TestDecoratorSignals(TestCase):
    """
    Make sure that the signals of the decorated models
    are processed correctly
    """

    def setUp(self):  # type () -> None
        """Setup the testing environment"""
        self.logged_instance = LoggedTestModel()
        self.other_instance = TestModel()

    def test_testmodels_exist_with_correct_values(self):  # type () -> None
        """Make sure all of the classes have the correct ground truth"""
        self.assertEqual(self.other_instance.name, 'TestModel')
        self.assertEqual(self.logged_instance.name, 'LoggedTestModel')

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_signals_on_logged_models(self):  # type () -> None
        """Test the 'create' and 'edit' signals"""

        # Mock task, as we dont want to send an actual log
        with patch('auditlogger.tasks.log_from_cloudwatch.delay') as mock_task:

            # Create
            self.logged_instance.save()
            # self.assertTrue(self.signal_was_called)
            self.assertTrue(mock_task.called)

            # Reset signal detector and test edit
            self.signal_was_called = False
            self.logged_instance.name = 'ChangedName'
            self.logged_instance.save()
            # self.assertTrue(self.signal_was_called)
            self.assertTrue(mock_task.called)

            # Reset signal detector and test delete
            self.signal_was_called = False
            self.logged_instance.delete()
            # self.assertTrue(self.signal_was_called)
            self.assertTrue(mock_task.called)

    def test_signals_on_models_that_are_not_logged(self):  # type () -> None
        """
        Test that models that are not being logged don't emmit any signals
        """

        # Mock task, as we dont want to send an actual log
        with patch('auditlogger.tasks.log_from_cloudwatch.delay') as mock_task:

            # Create normal instance, assert signal was not called
            self.other_instance.save()
            self.assertFalse(mock_task.called)


class TestUserSignals(TestCase):
    """
    Test that manipulation of the default User object triggers log actions
    """

    def setUp(self):  # type () -> None
        """Setup testing environment"""
        self.user = User.objects.create_user(
            'TestUser', 'test@example.com', 123,
        )

    def test_altering_user_fields_triggers_signals(self):
        """
        Make sure that changes to objects using the default User model
        trigger log actions
        """
        with patch('auditlogger.tasks.log_from_cloudwatch.delay') as mock_task:  # noqa
                # type () -> None
            self.client.login(username='TestUser', password='123')
            self.assertTrue(mock_task.called)



"""Signals are loaded"""
from django.apps import AppConfig  # type: ignore


class AuditLoggerConfig(AppConfig):
    """Set name and load signals"""
    name = 'auditlogger'

    def ready(self):
        """Load signals"""
        from auditlogger import signals  # noqa



"""Signals"""
import datetime
import inspect
import time

import django  # type: ignore
from django.contrib.auth.signals import user_logged_in  # type: ignore
from django.db.models.signals import post_delete, post_save  # type: ignore
from django.dispatch import receiver  # type: ignore
from six import string_types

import app_settings as app_settings
from middleware import AuditLoggerMiddleware
from tasks import log_to_redis


def get_client_ip(request):  # type () -> str
    """Fetch and return IP address of the user"""
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


@receiver(user_logged_in)
def user_logged_in(**kwargs):  # type () -> None
    """Detect a user login"""
    user = kwargs.pop('user')
    send_log('USER_LOGIN', user=user, **kwargs)


# The code below wont work on Django 1.4
if not django.get_version().startswith('1.4'):

    from django.contrib.auth.signals import user_login_failed

    @receiver(user_login_failed)
    def user_login_failed(**kwargs):
        try:
            user = kwargs.get('credentials', {}).get('username')
        except KeyError:
            user = None

        send_log('USER_LOGIN_FAILED', user=user, **kwargs)


def log_save(**kwargs):  # type () -> None
    """
    Detect if save is applied to a newly created object
    or if its an update. Log it to amazon cloudwatch.
    """
    action = 'OBJ_UPDATE' if not kwargs.get('created', False) else 'OBJ_CREATE'
    send_log(action, **kwargs)


def log_delete(**kwargs):  # type () -> None
    """Log delete"""
    send_log(action='OBJ_DELETE', **kwargs)


def send_log(action, request=None, user=None, **kwargs):  # type () -> None
    """Sends a log to amazon cloudwatch"""

    if not request:
        request = AuditLoggerMiddleware.get_request()

    # Instance data
    sender = kwargs.get('sender')
    sender_name = sender if isinstance(sender, string_types) else str(
        sender.__name__,
    )

    # Don't log excluded `objects
    if sender_name in app_settings.AUDIT_LOGGER_EXCLUDE:
        return

    # Create a list of updated fields, or return an empty one
    update_fields = kwargs.get('update_fields')
    fields = list(update_fields) if update_fields else []

    # Format instance data
    instance = kwargs.get('instance')
    instance_data = dict(
        model=sender_name,
        name=str(instance),
        pk=instance.pk,
        fields=fields,
        timestamp=int(time.time()) * 1000,
    ) if instance else None

    # Request data
    if hasattr(request, 'id'):
        request_id = getattr(request, 'id') if request else None
        url = request.build_absolute_uri() if request else None
        request_data = dict(id=request_id, url=url)
    else:
        request_data = ''

    # User data
    if not user and request:
        user = getattr(request, 'user')

    # Get user IP
    ip = get_client_ip(request) if request else None
    user_data = dict(ip=ip)

    if isinstance(user, string_types):
        user_data.update(username=user)
    elif user and user.is_authenticated():
        user_data.update(
            pk=user.pk,
            username=user.username,
            email=getattr(user, 'email'),
            is_staff=user.is_staff,
            is_superuser=user.is_superuser,
        )

    # Stack trace
    stack = ['{}:{} {}'.format(s[1], s[2], s[3]) for s in inspect.stack()]

    # Create date
    date = datetime.datetime.utcnow().isoformat()

    # Log data
    log_data = dict(
        date=date,
        request=request_data,
        action=action,
        user=user_data,
        instance=instance_data,
        stack=stack,
    )
    log_to_redis.delay(**log_data)


def auditlogger_update_delete(cls):  # type () -> Object
    """Decorator to connect the class to the correct signals"""

    post_save.connect(
        log_save, cls, dispatch_uid=cls.__name__,
    )
    post_delete.connect(
        log_delete, cls, dispatch_uid=cls.__name__,
    )
    return cls



"""Custom middleware for the auditlogger app"""
import threading
import uuid


class AuditLoggerMiddleware(object):
    """Custom middleware to fetch request data"""
    # Save current request
    _request = {}  # type: ignore

    def process_request(self, request):  # type () -> None
        """
        Add unique id to request, this way we can group them later.
        """
        request.id = uuid.uuid4().hex
        self.__class__.set_request(request)

    def process_response(self, _, response):  # type () -> None
        """Delete request"""
        self.__class__.del_request()
        return response

    def process_exception(self, _, __):  # type () -> None
        """Delete request"""
        self.__class__.del_request()

    @classmethod
    def get_request(cls, default=None):
        """Retrieve request"""
        return cls._request.get(threading.current_thread(), default)

    @classmethod
    def set_request(cls, request):
        """Store request"""
        cls._request[threading.current_thread()] = request

    @classmethod
    def del_request(cls):
        """Delete request"""
        cls._request.pop(threading.current_thread(), None)



""" Settings file to run stand alone tests"""
import django

SECRET_KEY = 'psst'
SITE_ID = 1

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    },
}

ROOT_URLCONF = 'auditlogger.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Middleware has changed between Django versions
if not django.get_version().startswith('1.4'):

    MIDDLEWARE = (
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    )

else:
    MIDDLEWARE_CLASSES = (
        'django.middleware.common.CommonMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
    )

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'auditlogger',
)



"""Classes to publish logs to aws"""
import json
import time

import boto3  # type: ignore
from botocore.config import Config  # type: ignore
from botocore.exceptions import ClientError  # type: ignore

import app_settings as app_settings  # type: ignore

# Limit the number of retries to 3, to stay under the rate limit
CONFIG = Config(
    retries=dict(
        max_attempts=3,
    ),
)


class AuditLoggerException(Exception):
    """Custom exception for the application"""
    pass


class CloudWatchLog:
    """Facilitates the connection to AWS"""
    next_sequence_token = None

    def __init__(self):  # type () -> None
        self.validate_settings()

        self.client = boto3.client(
            'logs',
            region_name=app_settings.AUDIT_LOGGER_REGION,
            aws_access_key_id=app_settings.AUDIT_LOGGER_KEY_ID,
            aws_secret_access_key=app_settings.AUDIT_LOGGER_SECRET_ACCESS_KEY,
            config=CONFIG,
        )

    @staticmethod
    def validate_settings():  # type () -> None
        """Make sure that the correct settings exist"""

        required_settings = [
            'AUDIT_LOGGER_REGION',
            'AUDIT_LOGGER_KEY_ID',
            'AUDIT_LOGGER_SECRET_ACCESS_KEY',
            'AUDIT_LOGGER_GROUP_NAME',
            'AUDIT_LOGGER_STREAM_NAME',
        ]

        for setting_name in required_settings:
            if not getattr(app_settings, setting_name):
                raise AuditLoggerException(
                    'Setting: {} not defined'.format(setting_name),
                )

    def create_log_group(self):  # type () -> None
        """Create (or skip) a loggroup"""
        # Try creating the log group
        try:
            self.client.create_log_group(
                logGroupName=app_settings.AUDIT_LOGGER_GROUP_NAME,
            )
        except ClientError:
            # The group probably already exists. We can just continue on
            # with the rest.
            pass

    def create_log_stream(self):  # type () -> None
        """Create (or skip) a logstream"""
        self.create_log_group()

        # Try creating the log stream
        try:
            self.client.create_log_stream(
                logGroupName=app_settings.AUDIT_LOGGER_GROUP_NAME,
                logStreamName=app_settings.AUDIT_LOGGER_STREAM_NAME,
            )
        except ClientError:
            # The stream probably already exists. We can just continue on
            # with the rest.
            pass

    def put_logs(self, log_data):  # type () -> None
        """Create new log objects and pass them to aws"""
        self.create_log_stream()

        log_events = []

        for log in log_data:
            entry = {
                'timestamp': int(time.time()) * 1000,
                'message': json.dumps(log),

            }
            log_events.append(entry)

        try:
            kwargs = dict(
                logGroupName=app_settings.AUDIT_LOGGER_GROUP_NAME,
                logStreamName=app_settings.AUDIT_LOGGER_STREAM_NAME,
                logEvents=log_events,
            )

            if self.next_sequence_token:
                kwargs['sequenceToken'] = self.next_sequence_token

            response = self.client.put_log_events(**kwargs)
            self.next_sequence_token = response.get('nextSequenceToken')

        except ClientError as e:
            if e.response.get('Error', {}).get('Code') in (
                    'DataAlreadyAcceptedException',
                    'InvalidSequenceTokenException'):
                self.next_sequence_token = \
                    e.response['Error']['Message'].rsplit(' ', 1)[-1]
                self.put_logs(log_data)
                return
            else:
                raise

        return



"""
setuptools config
"""
import setuptools

with open('README.md') as fh:
    long_description = fh.read()

setuptools.setup(
    name='django-audit-logger',
    version='1.0.2',
    author='Wend BV',
    author_email='info@wend.nl',
    description='A logger to be used internally',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='',
    packages=setuptools.find_packages(),
    classifiers=(
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ),
    install_requires=[
        'boto3',
        'celery',
        'mock;python_version=="2.7"',
        'django>=1.4',
        'redis',
    ],
)
